#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdbool.h>
#include <time.h>

struct team;
struct being;

struct skill {
    char *name;
    void (*skill)(struct being *, struct team *);
};

struct being{
    char *name;
    short max_health;
    short health;
    bool is_npc;
    bool is_final_boss;
    bool is_dead;
    short armour;
    short magic_resistance;
    short num_skills;
    struct skill *skills[];
};

struct team{
    short len;
    struct being *members[];
};

int read_int() {
    char buf[0x10] = {0};
    fgets(buf, sizeof(buf), stdin);
    return atoi(buf);
}

struct being* select_one(bool is_npc, struct team *target) {
    if (is_npc) {
        return target->members[rand() % target->len];
    }

    printf("available targets:\n");
    for (short i = 0; i < target->len; ++i) {
        printf("%2hd: %s\n", i, target->members[i]->name);
    }
    printf("> ");
    int choice = read_int();
    if (choice >= target->len) {
        printf("Your entire party is consumed by the eternal wraith of the DM. You cannot even choose a number from a given set! How sad.\n");
        exit(1);
    }
    return target->members[choice];
}

struct skill* select_skill(bool is_npc, struct being *caster) {
    if (is_npc) return caster->skills[rand() % caster->num_skills];
 
    printf("%ss turn:\n", caster->name);   
    printf("available skills:\n");
    for (short i = 0; i < caster->num_skills; ++i) {
        printf("%hd: %s\n", i, caster->skills[i]->name);
    }
    printf("> ");
    int choice = read_int();
    if (choice < 0 || choice >= caster->num_skills) {
        printf("Your entire party is consumed by the eternal wraith of the DM. You cannot even choose a number from a given set! How sad.\n");
        exit(1);
    }
    return caster->skills[choice];
}

void deal_damage(char *name, struct being *source, struct being *target, short amount) {
    if (target->is_dead)
        return printf("\x1b[31mThis action has been censored to achieve a family friendly rating by the BPjM\x1b[0m\n");
    if (target->is_final_boss && amount >= target->health)
        return printf("\x1b[31m%s\x1b[0m does not care about \x1b[32m%s\x1b[0ms damage!\n", target->name, source->name);
    if (amount == 0)
        return printf("There was an attempt .. to deal damage to \x1b[31m%s\x1b[0m by \x1b[32m%s\x1b[0m using \x1b[36m%s\x1b[0m\n", target->name, source->name, name);
    target->health -= amount;
    printf("\x1b[32m%s\x1b[0m deals \x1b[33m%hd\x1b[0m damage to \x1b[31m%s\x1b[0m using \x1b[36m%s\x1b[0m\n",source->name, amount, target->name, name);
}

void cleanup_team(struct team *target) {
    for (short i = 0; i < target->len; ++i) {
        struct being *t = target->members[i];
        if (!t->is_dead && t->health <= 0) {
            printf("\x1b[31m%s dies\x1b[0m\n", t->name);
            t->is_dead = true;
        }
    }
}

void skill_sword(struct being *caster, struct team *target) {
    struct being *t = select_one(caster->is_npc, target);
    short dmg = rand() & 0xf;
    deal_damage("sword slash", caster, t, dmg - t->armour > 0 ? dmg - t->armour : 0);
}

void skill_burn(struct being *caster, struct team *target) {
    short dmg = rand() & 7;
    for (short i = 0; i < target->len; ++i) {
        struct being *t = target->members[i];
        deal_damage("burning ball", caster, t, dmg - t->magic_resistance > 0 ? dmg - t->magic_resistance : 0);
    }
}

void skill_magic(struct being *caster, struct team *target) {
    struct being *t = select_one(caster->is_npc, target);
    short dmg = rand() % 0xc;
    deal_damage("magic missile", caster, t, dmg - t->magic_resistance > 0 ? dmg - t->magic_resistance : 0);
}

void skill_sacrifice(struct being *caster, struct team *target) {
    printf("how much life force does %s want to sacrifice (max %hd): ", caster->name, caster->health - 1);
    unsigned short dmg = read_int();
    unsigned short ch = caster->health;
    if (dmg >= ch) {
        printf("Your entire party is consumed by the eternal wraith of the DM. You cannot even choose a number from a given set! How sad.\n");
        exit(1);
    }
    deal_damage("sacrifice self", caster, caster, dmg);
    deal_damage("sacrifice self", caster, select_one(caster->is_npc, target), dmg);
}

struct skill sk_sw = {"sword slash", skill_sword};
struct skill sk_bu = {"burning ball", skill_burn};
struct skill sk_ma = {"magic missile", skill_magic};
struct skill sk_sa = {"sacrifice self", skill_sacrifice};
struct being baneful_balrog = {"baneful Balrog", 10000, 60, true, true, false, 10, 6, 2, {&sk_bu, &sk_ma}};
struct being wicked_wivern = {"wicked Wyvern", 2500, 2500, true, false, false, 15, 4, 1, {&sk_sw}};
struct being deadly_dragon = {"devastating Dragon", 4000, 4000, true, false, false, 5, 11, 1, {&sk_ma}};
struct being wimpy_wizzard = {"wimpy Wizard", 30, 30, false, false, false, 1, 3, 2, {&sk_ma, &sk_bu}};
struct being fragile_fighter = {"fragile Fighter", 50, 50, false, false, false, 7, 1, 2, {&sk_sw, &sk_sa}};
struct team accursed_adventurers = {2, {&wimpy_wizzard, &fragile_fighter}};
struct team eradicating_encounter = {3, {&baneful_balrog, &wicked_wivern, &deadly_dragon}};

bool is_finished() {
    if (baneful_balrog.is_dead) {
        printf("As you defeat the \"%s\" in this totally rigged encounter the DM glances at you with a smug smile on his face as he says:\n", baneful_balrog.name);
        printf("\tCongratulations on winning this exceptional encounter and killing my beloved beasts.\n");
        printf("\tWith \"%s\" defeated his cherished companions stagger and flee.\n", baneful_balrog.name);
        printf("\tAnd the prevailing party starts to loot the luxurious lair at once.\n");
        printf("\tBut the glamours gold soon turns out to be mischievous magic and\n");
        printf("\tall but a single item is nothing but a sinister scam.\n");
        printf("\tUpon close inspection you realize that even that last item is just a bogus banner\n");
        printf("\tnot knowing what you should do with \x1b[32m");
        system("get_flag");
        printf("\x1b[0m\tyou leave it where it is ragingly returning to the \"meticulous mayor\"\n");
        printf("\nYou should not have survived that... This striking story was to be continued in the devouring death realm\n");
        return true;
    }
    for (short i = 0; i < accursed_adventurers.len; ++i) {
        if (!accursed_adventurers.members[i]->is_dead) return false;
    }
    printf("The perishing party has been scattered by the ferocious forces of the evil.\n");
    printf("But much to your astonishment you find yourselves splashed on the shore of an particular peninsula...\n");
    return true;
}

int main(void) {
    setvbuf(stdout, NULL, _IONBF, 0);
    setvbuf(stdin, NULL, _IONBF, 0);
    srand(time(NULL));

    printf("You came together with some friends to play a role playing game, the DM praised to be \"extraordinarily exciting\"\n");
    printf("After running some errands for the \"meticulous mayor\" you make your way to the \"anticipated apex\" of the evening\n");
    do {
        printf("\n");
        for (short i = 0; i < eradicating_encounter.len; ++i) {
            if (eradicating_encounter.members[i]->is_dead) continue;
            struct skill *s = select_skill(true, eradicating_encounter.members[i]);
            s->skill(eradicating_encounter.members[i], &accursed_adventurers);
        }
        cleanup_team(&accursed_adventurers);
        for (short i = 0; i < accursed_adventurers.len; ++i) {
            if (accursed_adventurers.members[i]->is_dead) continue;
            printf("\n");
            struct skill *s = select_skill(false, accursed_adventurers.members[i]);
            s->skill(accursed_adventurers.members[i], &eradicating_encounter);
        }
        cleanup_team(&eradicating_encounter);
    } while(!is_finished());
    return 0;
}
